<?php

namespace Drupal\document_ocr\Entity;

use Drupal\Core\Field\BaseFieldDefinition;
use Drupal\Core\Entity\ContentEntityBase;
use Drupal\Core\Entity\EntityTypeInterface;
use Drupal\Core\Entity\ContentEntityInterface;
use Drupal\Core\Queue\DatabaseQueue;
use Drupal\file\Entity\File;

/**
 * Defines the Document OCR Import entity.
 *
 * @ingroup document_ocr
 *
 * @ContentEntityType(
 *   id = "document_ocr_import",
 *   label = @Translation("Document OCR Import"),
 *   handlers = {
 *     "view_builder" = "Drupal\Core\Entity\EntityViewBuilder",
 *     "list_builder" = "Drupal\document_ocr\Controller\ImportListBuilder",
 *     "views_data" = "Drupal\document_ocr\Entity\ViewsData\Import",
 *     "form" = {
 *       "process" = "\Drupal\document_ocr\Entity\Form\Import\ProcessForm",
 *       "entities" = "\Drupal\document_ocr\Entity\Form\Import\OlderEntitiesForm",
 *     },
 *   },
 *   base_table = "document_ocr_import",
 *   admin_permission = "administer document ocr",
 *   entity_keys = {
 *     "id" = "id",
 *     "label" = "id",
 *     "uuid" = "uuid",
 *   },
 *   links = {
 *     "collection" = "/admin/config/structure/document-ocr/imports",
 *     "process" = "/admin/config/structure/document-ocr/imports/process",
 *     "entities" = "/admin/config/structure/document-ocr/imports/entities",
 *   },
 * )
 */
class Import extends ContentEntityBase implements ContentEntityInterface {

  /**
   * {@inheritdoc}
   */
  public static function baseFieldDefinitions(EntityTypeInterface $entity_type) {
    $fields = parent::baseFieldDefinitions($entity_type);

    $fields['document_ocr_mapping'] = BaseFieldDefinition::create('entity_reference')
      ->setLabel(t('Document OCR'))
      ->setDescription(t('Document OCR Mapping entity reference.'))
      ->setSetting('target_type', 'document_ocr_mapping')
      ->setSetting('handler', 'default')
      ->setDisplayConfigurable('form', FALSE)
      ->setDisplayConfigurable('view', FALSE);

    $fields['source_entity_type_id'] = BaseFieldDefinition::create('string')
      ->setLabel(t('Source Entity type ID'))
      ->setSettings([
        'max_length' => 255,
        'text_processing' => 0,
      ])
      ->setDisplayConfigurable('form', FALSE)
      ->setDisplayConfigurable('view', FALSE);

    $fields['source_bundle'] = BaseFieldDefinition::create('string')
      ->setLabel(t('Source Bundle'))
      ->setSettings([
        'max_length' => 255,
        'text_processing' => 0,
      ])
      ->setDisplayConfigurable('form', FALSE)
      ->setDisplayConfigurable('view', FALSE);

    $fields['source_entity_id'] = BaseFieldDefinition::create('integer')
      ->setLabel(t('Source Entity ID'))
      ->setDescription(t('Imported entity ID.'))
      ->setDisplayConfigurable('form', FALSE)
      ->setDisplayConfigurable('view', FALSE);

    $fields['source_entity_revision_id'] = BaseFieldDefinition::create('integer')
      ->setLabel(t('Source Entity Revision ID'))
      ->setDescription(t('Imported entity revision ID.'))
      ->setDisplayConfigurable('form', FALSE)
      ->setDisplayConfigurable('view', FALSE);

    $fields['field_name'] = BaseFieldDefinition::create('string')
      ->setLabel(t('Field Name'))
      ->setDescription(t('Document OCR Mapping field name.'))
      ->setSettings([
        'max_length' => 255,
        'text_processing' => 0,
      ])
      ->setDisplayConfigurable('form', FALSE)
      ->setDisplayConfigurable('view', FALSE);

    $fields['file'] = BaseFieldDefinition::create('entity_reference')
      ->setLabel(t('Document File ID'))
      ->setDescription(t('Document File entity reference.'))
      ->setSetting('target_type', 'file')
      ->setSetting('handler', 'default')
      ->setDisplayConfigurable('form', FALSE)
      ->setDisplayConfigurable('view', FALSE);

    $fields['destination_entity_type_id'] = BaseFieldDefinition::create('string')
      ->setLabel(t('Destination Entity type ID'))
      ->setSettings([
        'max_length' => 255,
        'text_processing' => 0,
      ])
      ->setDisplayConfigurable('form', FALSE)
      ->setDisplayConfigurable('view', FALSE);

    $fields['destination_bundle'] = BaseFieldDefinition::create('string')
      ->setLabel(t('Destination Bundle'))
      ->setSettings([
        'max_length' => 255,
        'text_processing' => 0,
      ])
      ->setDisplayConfigurable('form', FALSE)
      ->setDisplayConfigurable('view', FALSE);

    $fields['destination_entity_id'] = BaseFieldDefinition::create('integer')
      ->setLabel(t('Destination Entity ID'))
      ->setDisplayConfigurable('form', FALSE)
      ->setDisplayConfigurable('view', FALSE);

    $fields['processed'] = BaseFieldDefinition::create('boolean')
      ->setLabel(t('Status'))
      ->setDescription(t('Indicating whether the document is processed or pending.'))
      ->setSetting('on_label', t('Processed'))
      ->setSetting('off_label', t('Pending'))
      ->setDefaultValue(FALSE)
      ->setDisplayConfigurable('form', FALSE)
      ->setDisplayConfigurable('view', FALSE);

    $fields['attempts'] = BaseFieldDefinition::create('integer')
      ->setLabel(t('Process Attempts'))
      ->setDescription(t('Number of process attempts.'))
      ->setDefaultValue(0)
      ->setDisplayConfigurable('form', FALSE)
      ->setDisplayConfigurable('view', FALSE);

    $fields['last_attempt'] = BaseFieldDefinition::create('integer')
      ->setLabel(t('Last Attempt'))
      ->setDescription(t('Last process attempt timestamp.'))
      ->setDisplayConfigurable('form', FALSE)
      ->setDisplayConfigurable('view', FALSE);

    return $fields;
  }

  /**
   * {@inheritdoc}
   */
  public function label() {
    if ($file = $this->getFile()) {
      return t('Import mapping for %filename (File ID: %fid)', [
        '%filename' => $file->getFilename(),
        '%fid' => $file->id()
      ]);
    }
    return parent::label();
  }

  /**
   * {@inheritdoc}
   */
  public function delete() {
    // Ceanup import item from the queue when import is deleted.
    $database = \Drupal::database();
    $query = $database->select(DatabaseQueue::TABLE_NAME, 't')
      ->condition('t.name', 'document_ocr')
      ->fields('t', ['item_id', 'data']);
    $queue = $query->execute();
    foreach ($queue as $record) {
      if ($item = unserialize($record->data)) {
        if (is_object($item) && $item->id == $this->id()) {
          $database->delete(DatabaseQueue::TABLE_NAME)
            ->condition('item_id', $record->item_id)
            ->execute();
        }
      }
    }
    parent::delete();
  }

  /**
   * {@inheritdoc}
   */
  public function attempt() {
    return $this->setAttempts($this->getAttempts() + 1);
  }

  /**
   * {@inheritdoc}
   */
  public function getAttempts() {
    return (int) $this->get('attempts')->value;
  }

  /**
   * {@inheritdoc}
   */
  public function setAttempts($attempts) {
    $this->set('attempts', (int) $attempts);
    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function resetAttempts() {
    return $this->setAttempts(0);
  }

  /**
   * {@inheritdoc}
   */
  public function getLastAttemptTime() {
    return $this->get('last_attempt')->value;
  }

  /**
   * {@inheritdoc}
   */
  public function setLastAttemptTime($timestamp = NULL) {
    $this->set('last_attempt', !empty($timestamp) ? $timestamp : time());
    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function getDocumentOcr() {
    if (!$this->get('document_ocr_mapping')->isEmpty()) {
      if ($entity = $this->get('document_ocr_mapping')->first()->get('entity')->getTarget()) {
        return $entity->getValue();
      }
    }
  }

  /**
   * {@inheritdoc}
   */
  public function setDocumentOcr($document_ocr_mapping) {
    $this->set('document_ocr_mapping', $document_ocr_mapping);
    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function getFile() {
    if (!$this->get('file')->isEmpty()) {
      if ($entity = $this->get('file')->first()->get('entity')->getTarget()) {
        return $entity->getValue();
      }
    }
  }

  /**
   * {@inheritdoc}
   */
  public function attachFile(File $file) {
    $this->set('file', $file->id());
    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function setProcessed() {
    $this->set('processed', TRUE);
    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function setPending() {
    $this->set('processed', FALSE);
    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function isProcessed() {
    return !empty($this->get('processed')->value);
  }

  /**
   * {@inheritdoc}
   */
  public function isPending() {
    return empty($this->get('processed')->value);
  }

  /**
   * {@inheritdoc}
   */
  public function attachSource($entity) {
    $this->set('source_entity_type_id', $entity->getEntityTypeId());
    $this->set('source_bundle', $entity->bundle());
    $this->set('source_entity_id', $entity->id());
    $this->set('source_entity_revision_id', $entity->getRevisionId());
    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function getSourceEntity() {
    if (!$this->get('source_entity_type_id')->isEmpty()) {
      if ($entity = $this->entityTypeManager()
          ->getStorage($this->get('source_entity_type_id')->value)
          ->load($this->get('source_entity_id')->value)) {
        return $entity;
      }
    }
  }

  /**
   * {@inheritdoc}
   */
  public function attachField($field_name) {
    $this->set('field_name', $field_name);
    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function attachDestination($entity) {
    $this->set('destination_entity_type_id', $entity->getEntityTypeId());
    $this->set('destination_bundle', $entity->bundle());
    $this->set('destination_entity_id', $entity->id());
    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function getDestinationEntity() {
    if (!$this->get('destination_entity_type_id')->isEmpty()) {
      if ($entity = $this->entityTypeManager()
          ->getStorage($this->get('destination_entity_type_id')->value)
          ->load($this->get('destination_entity_id')->value)) {
        return $entity;
      }
    }
  }

}
