<?php

namespace Drupal\document_ocr\Plugin\document_ocr\transformer;

use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\File\FileSystemInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Drupal\Component\Utility\Unicode;
use Drupal\document_ocr\Plugin\TransformerBase;
use Drupal\document_ocr\Services\OpenAi as OpenAiService;

/**
 * Open AI integration transformer plugin.
 *
 * @DocumentOcrTransformer(
 *   id = "openai",
 *   name = @Translation("OpenAI"),
 *   description = @Translation("OpenAI integration transformer."),
 *   require_credentials = true
 * )
 */
class OpenAi extends TransformerBase {

  /**
   * OpenAI serivce.
   *
   * @var \Drupal\document_ocr\Services\OpenAi as OpenAiService
   */
  protected $openAi;

  /**
   * {@inheritdoc}
   */
  public function __construct(array $configuration, $plugin_id, $plugin_definition, EntityTypeManagerInterface $entity_type_manager, FileSystemInterface $file_system, OpenAiService $openai) {
    parent::__construct($configuration, $plugin_id, $plugin_definition, $entity_type_manager, $file_system);
    $this->openAi = $openai;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static($configuration, $plugin_id, $plugin_definition,
      $container->get('entity_type.manager'),
      $container->get('file_system'),
      $container->get('document_ocr.openai')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function configurationForm(array $form, FormStateInterface $form_state) {
    $this->openAi->setCredentials($this->getCredentials());
    $configuration = $this->getConfiguration();
    $form['prompt'] = [
      '#type' => 'textarea',
      '#title' => $this->t('Prompt'),
      '#description' => $this->t('Provide OpenAI command and use <strong>%text%</strong> for the document content, or use <strong>%text:600%</strong> to limit text to 600 characters'),
      '#default_value' => !empty($configuration['prompt']) ? $configuration['prompt'] : '',
      '#required' => TRUE,
    ];
    $form['model'] = [
      '#type' => 'select',
      '#title' => $this->t('Model'),
      '#description' => $this->t('The model which will generate the completion. Some models are suitable for natural language tasks, others specialize in code. <a href="@link" target="_blank">Learn more</a>.', ['@link' => 'https://platform.openai.com/docs/models']),
      '#options' => $this->openAi->getModels(),
      '#default_value' => !empty($configuration['model']) ? $configuration['model'] : 'text-davinci-003',
      '#required' => TRUE,
    ];
    $form['temperature'] = [
      '#type' => 'number',
      '#title' => $this->t('Temperature'),
      '#description' => $this->t('Controls randomness: Lowering results in less random completions. As the temperature approaches zero, the model will become deterministic and repetitive.'),
      '#min' => 0,
      '#max' => 2,
      '#step' => 0.01,
      '#default_value' => !empty($configuration['temperature']) ? $configuration['temperature'] : 0.7,
    ];
    $form['max_tokens'] = [
      '#type' => 'number',
      '#title' => $this->t('Maximum length'),
      '#description' => $this->t('The maximum number of tokens to generate. Requests can use up to 2,048 or 4,000 tokens shared between prompt and completion. The exact limit varies by model. (One token is roughly 4 characters for normal English text)'),
      '#min' => 1,
      '#max' => 4000,
      '#default_value' => !empty($configuration['max_tokens']) ? $configuration['max_tokens'] : 60,
    ];
    $form['top_p'] = [
      '#type' => 'number',
      '#title' => $this->t('Top P'),
      '#description' => $this->t('Controls diversity via nucleus sampling: 0.5 means half of all likelihood-weighted options are considered.'),
      '#min' => 0,
      '#max' => 1,
      '#step' => 0.01,
      '#default_value' => !empty($configuration['top_p']) ? $configuration['top_p'] : 1.0,
    ];
    $form['frequency_penalty'] = [
      '#type' => 'number',
      '#title' => $this->t('Frequency penalty'),
      '#description' => $this->t("How much to penalize new tokens based on their existing frequency in the text so far. Decreases the model's likelihood to repeat the same line verbatim."),
      '#min' => 0,
      '#max' => 2,
      '#step' => 0.01,
      '#default_value' => !empty($configuration['frequency_penalty']) ? $configuration['frequency_penalty'] : 0.0,
    ];
    $form['presence_penalty'] = [
      '#type' => 'number',
      '#title' => $this->t('Presence penalty'),
      '#description' => $this->t("How much to penalize new tokens based on whether they appear in the text so far. Increases the model's likelihood to talk about new topics."),
      '#min' => 0,
      '#max' => 2,
      '#step' => 0.01,
      '#default_value' => !empty($configuration['presence_penalty']) ? $configuration['presence_penalty'] : 1,
    ];
    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function configurationValues(array &$form, FormStateInterface $form_state) {
    return [
      'prompt' => $form_state->getValue('prompt'),
      'model' => (string) $form_state->getValue('model'),
      'temperature' => (float) $form_state->getValue('temperature'),
      'max_tokens' => (int) $form_state->getValue('max_tokens'),
      'top_p' => (float) $form_state->getValue('top_p'),
      'frequency_penalty' => (float) $form_state->getValue('frequency_penalty'),
      'presence_penalty' => (float) $form_state->getValue('presence_penalty'),
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function transform($value, $configuration) {
    try {
      $this->openAi->setCredentials($this->getCredentials());
      if (preg_match('/%text:(.*)%/', $configuration['prompt'], $match)) {
        if (!empty($match[1])) {
          $configuration['prompt'] = str_replace($match[0], Unicode::truncate($value, (int) $match[1], TRUE), $configuration['prompt']);
        }
      }
      else {
        $configuration['prompt'] = str_replace('%text%', $value, $configuration['prompt']);
      }
      if ($response = $this->openAi->request($configuration)) {
        if (!empty($response->choices[0])) {
          $choice = $response->choices[0]->toArray();
          return $choice['text'];
        }
      }
    }
    catch (\Exception $ex) {
      return $value;
    }
  }

}
