<?php

namespace Drupal\document_ocr\Services;

use Drupal\Core\File\FileSystemInterface;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Drupal\Core\Cache\CacheBackendInterface;
use Drupal\Core\Datetime\DateFormatterInterface;
use Google\ApiCore\ApiException;
use Google\Cloud\DocumentAI\V1\DocumentProcessorServiceClient;
use Google\Cloud\DocumentAI\V1\RawDocument;
use Google\Cloud\DocumentAI\V1\ProcessorVersion\State;

/**
 * Google Document AI service.
 */
class GoogleDocumentAi {

  use StringTranslationTrait;

  /**
   * The file system service.
   *
   * @var \Drupal\Core\File\FileSystemInterface
   */
  protected $fileSystem;

  /**
   * The Google Document process service.
   *
   * @var \Google\Cloud\DocumentAI\V1\DocumentProcessorServiceClient
   */
  protected $document_processor;

  /**
   * The Google project name.
   *
   * @var string
   */
  protected $project_id;

  /**
   * The processor parameters.
   *
   * @var array
   */
  protected $processor_params;

  /**
   * The Google Service credentials.
   *
   * @var string
   */
  protected $credentials;

  /**
   * The cache.default cache backend.
   *
   * @var \Drupal\Core\Cache\CacheBackendInterface
   */
  protected $cacheBackend;

  /**
   * Cache expiration in seconds.
   */
  protected $interval = 600;

  /**
   * {@inheritdoc}
   */
  public function __construct(FileSystemInterface $file_system, CacheBackendInterface $cache) {
    $this->fileSystem = $file_system;
    $this->cacheBackend = $cache;
  }

  /**
   * Set JSON credentials file.
   */
  public function setCredentials($path) {
    // Create a client.
    $this->credentials = json_decode(file_get_contents(
        $this->fileSystem->realpath($path)
      ), TRUE);
    $this->setProjectId($this->credentials['project_id']);
    return $this;
  }

  public function getVersionStateName($state) {
    return State::name($state);
  }

  /**
   * Initialize servie client.
   */
  public function serviceClient($location_id = 'us') {
    $params = [
      'credentials' => $this->credentials,
    ];
    if ($location_id != 'us') {
      $params['apiEndpoint'] = $location_id . '-documentai.googleapis.com';
    }
    $this->document_processor = new DocumentProcessorServiceClient($params);
    return $this;
  }

  /**
   * Get a project ID.
   */
  public function getProjectId() {
    return $this->project_id;
  }

  /**
   * Set a project ID.
   */
  public function setProjectId($id) {
    $this->project_id = $id;
    return $this;
  }

  /**
   * Get a processor name.
   */
  public function getParams() {
    return $this->processor_params;
  }

  /**
   * Set a processor name.
   */
  public function setParams($processor_name) {
    $this->processor_params = $this->document_processor->parseName($processor_name);
    return $this;
  }

  /**
   * Get processors.
   */
  public function getProcessors($location_id) {
    $processors = [];
    $cid = 'google_document_ai:processors:' . $this->getProjectId() . ':' . $location_id;
    if ($cache = $this->cacheBackend->get($cid)) {
      $processors = $cache->data;
    }
    else {
      $this->serviceClient($location_id);
      $formatted_parent = $this->document_processor->locationName($this->getProjectId(), $location_id);
      try {
        $response = $this->document_processor->listProcessors($formatted_parent);
        foreach ($response as $element) {
          $processors[$element->getName()] = $element;
        }
      }
      catch (ApiException $ex) {
        \Drupal::logger('document_ocr')->error($ex->getMessage());
      }
      $this->cacheBackend->set($cid, $processors, time() + $this->interval, [
        'google_document_ai:processors',
        'google_document_ai:processors:' . $location_id
      ]);
    }
    return $processors;
  }

  /**
   * Get processors versions.
   */
  public function getProcessorVersions() {
    $params = $this->getParams();
    $location_id = $params['location'];
    $processor_id = $params['processor'];
    $versions = [];
    $cid = 'google_document_ai:processor_version:' . $this->getProjectId() . ':' . $processor_id;
    if ($cache = $this->cacheBackend->get($cid)) {
      $versions = $cache->data;
    }
    else {
      $this->serviceClient($location_id);
      $formatted_parent = $this->document_processor->processorName($params['project'], $location_id, $processor_id);
      try {
        $response = $this->document_processor->listProcessorVersions($formatted_parent);
        foreach ($response as $element) {
          $versions[$element->getName()] = $element;
        }
      }
      catch (ApiException $ex) {
        \Drupal::logger('document_ocr')->error($ex->getMessage());
      }
      $this->cacheBackend->set($cid, $versions, time() + $this->interval, [
        'google_document_ai:processors',
        'google_document_ai:processor_version:' . $processor_id
      ]);
    }
    return $versions;
  }

  /**
   * Get locations.
   */
  public function getLocations() {
    $locations = [];
    $cid = 'google_document_ai:locations:' . $this->getProjectId();
    if ($cache = $this->cacheBackend->get($cid)) {
      $locations = $cache->data;
    }
    else {
      try {
        $response = $this->document_processor->listLocations([
          'name' => 'projects/' . $this->getProjectId(),
        ]);
        foreach ($response as $element) {
          $locations[$element->getLocationId()] = $this->regionNames($element->getLocationId());
        }
      }
      catch (ApiException $ex) {
        \Drupal::logger('document_ocr')->error($ex->getMessage());
      }
      $this->cacheBackend->set($cid, $locations, time() + $this->interval, [
        'google_document_ai:locations',
      ]);
    }
    return $locations;
  }

  /**
   * Process document.
   */
  public function processDocument($file) {
    try {
      $params = $this->getParams();
      $this->serviceClient($params['location']);
      $documentPath = $this->fileSystem->realpath($file->getFileUri());

      $handle = fopen($documentPath, 'rb');
      $contents = fread($handle, filesize($documentPath));
      fclose($handle);

      $rawDocument = new RawDocument([
        'content' => $contents,
        'mime_type' => $file->getMimeType(),
      ]);

      if (!empty($params['processor_version'])) {
        $name = $this->document_processor->processorVersionName($this->getProjectId(), $params['location'], $params['processor'], $params['processor_version']);
      }
      else {
        $name = $this->document_processor->processorName($this->getProjectId(), $params['location'], $params['processor']);
      }
      $response = $this->document_processor->processDocument($name, [
        'rawDocument' => $rawDocument
      ]);
      return $response->getDocument();
    }
    catch (ApiException $ex) {
      \Drupal::logger('document_ocr')->error($ex->getMessage());
    }
  }

  /**
   * Region names.
   */
  protected function regionNames($location_id) {
    $locations = [
      'us' => $this->t('United States'),
      'eu' => $this->t('Europe'),
      'asia-south1' => $this->t('Mumbai'),
      'asia-southeast1' => $this->t('Singapore'),
      'australia-southeast1' => $this->t('Sydney'),
      'europe-west2' => $this->t('London'),
      'europe-west3' => $this->t('Frankfurt'),
      'northamerica-northeast1' => $this->t('Montréal'),
    ];
    return !empty($locations[$location_id]) ?$locations[$location_id] : strtoupper($location_id);
  }
}
