<?php

namespace Drupal\document_ocr\Services;

use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Entity\EntityFieldManager;
use Drupal\Core\Queue\QueueFactory;
use Drupal\Core\Queue\DatabaseQueue;
use Drupal\document_ocr\Plugin\ProcessorManager;
use Drupal\document_ocr\Plugin\TransformerManager;

/**
 * Document OCR process service.
 */
class Process {

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The entity field manager.
   *
   * @var \Drupal\Core\Entity\EntityFieldManager
   */
  protected $entityFieldManager;

  /**
   * Processor manager.
   *
   * @var \Drupal\document_ocr\Plugin\ProcessorManager
   */
  protected $processorManager;

  /**
   * Processor manager.
   *
   * @var \Drupal\document_ocr\Plugin\TransformerManager
   */
  protected $transformerManager;

  /**
   * The queue object.
   *
   * @var \Drupal\Core\Queue\QueueFactory
   */
  protected $queue;

  /**
   * Storage for the document_ocr_import entity.
   */
  protected $importStorage;

  /**
   * {@inheritdoc}
   */
  public function __construct(EntityTypeManagerInterface $entity_type_manager, EntityFieldManager $entity_field_manager, QueueFactory $queue, ProcessorManager $processor, TransformerManager $transformer) {
    $this->entityTypeManager = $entity_type_manager;
    $this->entityFieldManager = $entity_field_manager;
    $this->queue = $queue;
    $this->processorManager = $processor;
    $this->transformerManager = $transformer;
    $this->importStorage = $entity_type_manager->getStorage('document_ocr_import');
  }

  /**
   * Process entity file.
   */
  public function processEntity(EntityInterface $entity) {
    $documents = $this->matchDocumentMapping($entity);
    foreach ($documents as $document) {
      $field = $document->getFieldSelector()['field'];
      $settings = $document->getSettings();
      // Process only active mapping.
      if (!$document->isActive()) {
        continue;
      }
      if (!$entity->isNew() && empty($settings['process_update'])) {
        continue;
      }
      $processor = $document->getProcessorReference()->getProcessorPlugin();
      if (!$entity->get($field)->isEmpty()) {
        foreach ($entity->get($field)->referencedEntities() as $file) {
          // Only processing supported file extensions.
          if ($this->isExtensionSupported($file, $processor->getExtensions())) {
            $queue_item = TRUE;
            $import = $this->importReferenceEntity($document, $entity, $file);
            if (!empty($settings['realtime_processing'])) {
              if ($destination = $this->processDocument($document, $file)) {
                $import->attachFile($file)
                  ->attachField($field)
                  ->attachDestination($destination)
                  ->setProcessed()
                  ->save();
                $queue_item = FALSE;
              }
              else {
                $import->attempt()
                  ->setPending()
                  ->save();
              }
            }
            if ($queue_item) {
              $import->attachFile($file)
                ->attachField($field)
                ->setPending()
                ->save();
              $this->addQueueItem($import);
            }
          }
        }
      }
    }
  }

  /**
   * Process entity file.
   */
  public function deleteEntity(EntityInterface $entity) {
    if ($entity->getEntityTypeId() == 'file') {
      $imports = $this->importStorage->loadByProperties(['file' => $entity->id()]);
      foreach ($imports as $import) {
        // Deleting destination entity. @see Document OCR settings page.
        $document = $import->getDocumentOcr();
        $settings = $document->getSettings();
        if (!empty($settings['delete_destination'])) {
          $entities = $this->entityTypeManager->getStorage($import->get('entity_type_id')->value)->loadByProperties([
            'file' => $entity->id(),
          ]);
          if ($destination = reset($entities)) {
            $destination->delete();
          }
        }
        // Deleting import reference.
        $import->delete();
      }
    }
    // Delete multiple import references when source node is deleted.
    $source_entities = $this->importStorage->loadByProperties([
      'source_entity_type_id' => $entity->getEntityTypeId(),
      'source_bundle' => $entity->bundle(),
      'source_entity_id' => $entity->id(),
    ]);
    foreach ($source_entities as $source) {
      $source->delete();
    }
    // Delete multiple import references when imported node is deleted.
    $destination_entities = $this->importStorage->loadByProperties([
      'destination_entity_type_id' => $entity->getEntityTypeId(),
      'destination_bundle' => $entity->bundle(),
      'destination_entity_id' => $entity->id(),
    ]);
    foreach ($destination_entities as $destination) {
      $destination->delete();
    }
  }

  /**
   * Run pending imports if faild add back to the queue until attempt limit reached.
   */
  public function runPendingImport($import, $item) {
    $document = $import->getDocumentOcr();
    if (!$document->isActive()) {
      return;
    }
    if ($file = $import->getFile()) {
      if ($destination = $this->processDocument($document, $file)) {
        $import->attachDestination($destination)
          ->setProcessed()
          ->resetAttempts()
          ->save();
        $this->removeFromQueues($import);
        return $destination;
      }
      else {
        $import->attempt()
          ->setPending()
          ->save();
        // Add to retry queue if initial processing failed.
        if ($item->queue_id == 'document_ocr') {
          $this->addRetryQueueItem($import);
        }
        else {
          $this->addQueueItem($import);
        }
      }
    }
  }

  /**
   * Process older entity file.
   */
  public function prcoessOlderEntity($document, $entity) {
    if (!$document->isActive()) {
      return;
    }
    $field = $document->getFieldSelector()['field'];
    $settings = $document->getSettings();
    if (!$entity->get($field)->isEmpty()) {
      foreach ($entity->get($field)->referencedEntities() as $file) {
        $import = $this->importReferenceEntity($document, $entity, $file);
        $import->attachFile($file)
          ->attachField($field)
          ->setPending()
          ->save();
      }
      return $entity;
    }
  }

  /**
   * Process document, build destination entity using document properties and save it.
   */
  public function processDocument($document, $file) {
    $document_ocr_processor = $document->getProcessorReference();
    $processor = $document_ocr_processor->getProcessorPlugin();
    $processor->setCredentials($document_ocr_processor->getCredentials())
      ->setConfiguration($document_ocr_processor->getConfiguration())
      ->setFile($file);
    $destination = $document->getDestinationBundle();
    if ($mapping_data = $processor->getMappingData()) {
      $imports = $this->importStorage->loadByProperties([
        'document_ocr_mapping' => $document->id(),
        'destination_entity_type_id' => $destination['entity_type_id'],
        'destination_bundle' => $destination['bundle'],
        'file' => $file->id(),
      ]);
      $storage = $this->entityTypeManager->getStorage($destination['entity_type_id']);
      if ($import = reset($imports)) {
        $entity = $storage->load($import->get('destination_entity_id')->value);
      }
      else {
        $entity = $storage->create([
          'bundle' => $destination['bundle'],
          'type' => $destination['bundle'],
        ]);
      }
      return $this->saveEntity($entity, $document->getMapping(), $mapping_data, $file);
    }
  }

  /**
   * Run pending imports if faild add back to the queue until attempt limit reached.
   */
  public function runOneTimeImport($item) {
    $processor = $item->processor->getProcessorPlugin();
    $processor->setCredentials($item->processor->getCredentials())
      ->setConfiguration($item->processor->getConfiguration())
      ->setFile($item->file);
    if ($mapping_data = $processor->getMappingData()) {
      list ($entity_type_id, $bundle) = explode('::', $item->destination_bundle);
      $entity = $this->entityTypeManager->getStorage($entity_type_id)->create([
        'bundle' => $bundle,
        'type' => $bundle,
      ]);
      return $this->saveEntity($entity, $item->mapping, $mapping_data, $item->file, $item->field);
    }
  }

  /**
   * Map fields and document properties.
   */
  protected function saveEntity($entity, $mapping, $mapping_data, $file, $field = NULL) {
    foreach ($mapping as $field_name => $map) {
      if (!empty($map['property']) && !empty($mapping_data->getValues()[$map['property']])) {
        $field_value = $mapping_data->getValues()[$map['property']];
        if ($document_ocr_transformer = $this->entityTypeManager->getStorage('document_ocr_transformer')->load($map['transformer'])) {
          $transformer = $this->transformerManager->createInstance($document_ocr_transformer->getTransformer());
          $transformer->setCredentials($document_ocr_transformer->getCredentials())->setMappingData($mapping_data);
          $field_value = $transformer->transform($field_value, $document_ocr_transformer->getConfiguration());
        }
        $entity->set($field_name, $field_value);
      }
      else {
        if ($map['transformer'] == 'custom_value') {
          $entity->set($field_name, $map['custom_value']);
        }
      }
    }
    // If title is not empty set it to the filename. The title field is required and cannot be empty.
    if ($entity->hasField('title')) {
      if (empty($entity->get('title')->value)) {
        $entity->set('title', $file->getFilename());
      }
    }
    else if ($entity->hasField('label')) {
      if (empty($entity->get('label')->value)) {
        $entity->set('label', $file->getFilename());
      }
    }
    if (!empty($field)) {
      $entity->set($field, $file->id());
    }
    $entity->save();
    return $entity;
  }

  /**
   * Get entity mapping.
   */
  protected function matchDocumentMapping($entity) {
    $storage = $this->entityTypeManager->getStorage('document_ocr_mapping');
    $ids = $storage->getQuery()
      ->condition('field_selector.entity_type_id', $entity->getEntityTypeId())
      ->condition('field_selector.bundle', $entity->bundle())
      ->accessCheck(FALSE)
      ->execute();
    return $storage->loadMultiple($ids);
  }

  /**
   * Save entity mapping reference.
   */
  protected function importReferenceEntity($document, $entity, $file) {
    $existing = $this->importStorage->loadByProperties([
      'document_ocr_mapping' => $document->id(),
      'source_entity_type_id' => $entity->getEntityTypeId(),
      'source_bundle' => $entity->bundle(),
      'source_entity_id' => $entity->id(),
      'source_entity_revision_id' => $entity->getRevisionId(),
      'file' => $file->id(),
    ]);
    if (!$import = reset($existing)) {
      $import = $this->importStorage->create([]);
      $import->setDocumentOcr($document->id());
      $import->attachSource($entity);
    }
    $import->setLastAttemptTime();
    return $import;
  }

  /**
   * Add document to the queue.
   */
  protected function addQueueItem($import) {
    $query = \Drupal::database()->select(DatabaseQueue::TABLE_NAME, 't')
      ->condition('t.name', 'document_ocr')
      ->fields('t', ['item_id', 'data']);
    $queue = $query->execute();
    $queue_ids = [];
    foreach ($queue as $record) {
      if ($item = unserialize($record->data)) {
        if (is_object($item) && $item->id == $import->id()) {
          $queue_ids[] = $item->id;
        }
      }
    }
    // Do not add document to process queue if its already in it.
    if (!in_array($import->id(), $queue_ids)) {
      $item = new \stdClass();
      $item->id = $import->id();
      $this->queue->get('document_ocr')->createItem($item);
    }
  }

  /**
   * Add document to the retry queue.
   */
  protected function addRetryQueueItem($import) {
    $query = \Drupal::database()->select(DatabaseQueue::TABLE_NAME, 't')
      ->condition('t.name', 'document_ocr_retry')
      ->fields('t', ['item_id', 'data']);
    $queue = $query->execute();
    $queue_ids = [];
    foreach ($queue as $record) {
      if ($item = unserialize($record->data)) {
        if (is_object($item) && $item->id == $import->id()) {
          $queue_ids[] = $item->id;
        }
      }
    }
    // Do not add document to process queue if its already in it.
    if (!in_array($import->id(), $queue_ids)) {
      $item = new \stdClass();
      $item->id = $import->id();
      $this->queue->get('document_ocr_retry')->createItem($item);
    }
  }

  /**
   * Remove items from the queue.
   */
  protected function removeFromQueues($import) {
    $query = \Drupal::database()->select(DatabaseQueue::TABLE_NAME, 't')
      ->condition('t.name', ['document_ocr', 'document_ocr_retry'], 'IN')
      ->fields('t', ['item_id', 'data']);
    $queue = $query->execute();
    $queue_ids = [];
    foreach ($queue as $record) {
      if ($item = unserialize($record->data)) {
        if (is_object($item) && $item->id == $import->id()) {
          $queue_ids[] = $record;
        }
      }
    }
    if (!empty($queue_ids)) {
      foreach ($queue_ids as $item) {
        $this->queue->get('document_ocr')->deleteItem($item);
      }
    }
  }

  /**
   * Check if file extension supported.
   */
  protected function isExtensionSupported($file, $extensions) {
    $regex = '/\.(' . preg_replace('/ +/', '|', preg_quote($extensions)) . ')$/i';
    $subject = $file->isTemporary() ? $file->getFilename() : $file->getFileUri();
    return preg_match($regex, $subject);
  }

}
